#
# Copyright 2022 Amazon.com, Inc. or its affiliates. All Rights Reserved.
# Copyright (C) 2018-2022  OpenCyphal Development Team  <opencyphal.org>
# This software is distributed under the terms of the MIT License.
#
"""

.. autodata:: __version__
"""

__version__ = "2.0.5"
__license__ = "MIT"
__author__ = "OpenCyphal"
__copyright__ = "Copyright 2022 Amazon.com, Inc. or its affiliates. All Rights Reserved. Copyright (c) 2022 OpenCyphal."
__email__ = "maintainers@opencyphal.org"



#
# Copyright 2021 Amazon.com, Inc. or its affiliates. All Rights Reserved.
# Copyright (C) 2021  OpenCyphal Development Team  <opencyphal.org>
# This software is distributed under the terms of the MIT License.
#
"""
A small collection of common utilities.

.. note::

    Please don't use this as a dumping ground for things that belong in a dedicated package. Python being such a
    full-featured language, there should be very few truly generic utilities in Nunavut.

"""
import collections.abc
import copy
import enum
import logging
import pathlib
from typing import Generator, MutableMapping, cast, TypeVar

import importlib_resources

_logger = logging.getLogger(__name__)


TEMPLATE_SUFFIX = ".j2"  #: The suffix expected for nunavut templates.


@enum.unique
class YesNoDefault(enum.Enum):
    """
    Trinary type for decisions that allow a default behavior to be requested that can
    be different based on other contexts. For example:

    .. invisible-code-block: python

        from datetime import datetime
        from nunavut._utilities import YesNoDefault

    .. code-block:: python

        def should_we_order_pizza(answer: YesNoDefault) -> bool:
            if answer == YesNoDefault.YES or (
               answer == YesNoDefault.DEFAULT and
               datetime.today().isoweekday() == 5):
                # if yes or if we are taking the default action which is to
                # order pizza on Friday, and today is Friday, then we order pizza
                return True
            else:
                return False

    .. invisible-code-block: python

        assert should_we_order_pizza(YesNoDefault.YES)
        assert not should_we_order_pizza(YesNoDefault.NO)

    """

    @classmethod
    def test_truth(cls, ynd_value: "YesNoDefault", default_value: bool) -> bool:
        """
        Helper method to test a YesNoDefault value and return a default boolean value.

        .. invisible-code-block: python

            from nunavut._utilities import YesNoDefault

        .. code-block:: python

            '''
                let "is YES" be Y
                let "is DEFAULT" be D where:
                    if Y then not D and if D then not Y
                    and "is NO" is Y = D = 0
                let "is default_value true" be d

                Y | D | d | Y or (D and d)
                1   *   *    1
                0   1   0    0
                0   1   1    1
                0   0   *    0
            '''

            assert YesNoDefault.test_truth(YesNoDefault.YES, False)
            assert not YesNoDefault.test_truth(YesNoDefault.DEFAULT, False)
            assert YesNoDefault.test_truth(YesNoDefault.DEFAULT, True)
            assert not YesNoDefault.test_truth(YesNoDefault.NO, True)

        """
        if ynd_value == cls.DEFAULT:
            return default_value
        else:
            return ynd_value == cls.YES

    NO = 0
    YES = 1
    DEFAULT = 2


@enum.unique
class ResourceType(enum.Enum):
    """
    Common Nunavut classifications for Python package resources.
    """

    ANY = 0
    CONFIGURATION = 1
    SERIALIZATION_SUPPORT = 2
    TYPE_SUPPORT = 3


@enum.unique
class ResourceSearchPolicy(enum.Enum):
    """
    Generic policy type for controlling the behaviour of things that seach for resources.
    """

    FIND_ALL = 0
    FIND_FIRST = 1


def iter_package_resources(pkg_name: str, *suffix_filters: str) -> Generator[pathlib.Path, None, None]:
    """
    >>> from nunavut._utilities import iter_package_resources
    >>> rs = [x for x in iter_package_resources("nunavut.lang", ".py") if x.name == "__init__.py"]
    >>> len(rs)
    1
    >>> rs[0].name
    '__init__.py'

    """
    for resource in importlib_resources.files(pkg_name).iterdir():
        if resource.is_file() and isinstance(resource, pathlib.Path):
            # Not sure why this works but it's seemed to so far. importlib_resources.as_file(resource)
            # may be more correct but this can create temporary files which would disappear after the iterator
            # had copied their paths. If you are reading this because this method isn't working for some packaging
            # scheme then we may need to use importlib_resources.as_file(resource) to create a runtime cache of
            # temporary objects that live for a given nunavut session. This, of course, wouldn't help across sessions
            # which is a common use case when integrating Nunavut with build systems. So...here be dragons.
            file_resource = resource
            if any(suffix == file_resource.suffix for suffix in suffix_filters):
                yield file_resource


def empty_list_support_files() -> Generator[pathlib.Path, None, None]:
    """
    Helper for implementing the list_support_files method in language support packages. This provides an empty
    iterator with the correct type annotations.
    """
    # works in Python 3.3 and newer. Thanks https://stackoverflow.com/a/13243870
    yield from ()


DeepUpdateType = TypeVar("DeepUpdateType", bound=MutableMapping)


def deep_update(target: DeepUpdateType, source: DeepUpdateType) -> DeepUpdateType:
    """
    Helper method to do a recursive update of a map that may contain maps as values.

    .. invisible-code-block: python

        from nunavut._utilities import deep_update
        import collections.abc

    .. code-block:: python

        target_map   =  {
                            "a": { "one": 1, "two": 2 },
                            "b": "not a map"
                        }
        update_from  =  {
                            "a": { "two": { "i": "this value" }, "three": "that value"},
                            "c": "see"
                        }

        target_map   = deep_update(target_map, update_from)
        update_from["a"]["two"]["i"] = "whoops, this was supposed to be a copy"

        assert target_map["a"]["one"] == 1
        assert isinstance(target_map["a"]["two"], collections.abc.Mapping)
        assert target_map["a"]["two"]["i"] == "this value"
        assert target_map["a"]["three"] == "that value"
        assert target_map["b"] == "not a map"
        assert "c" in target_map
        assert target_map["c"] == "see"

    """
    if isinstance(target, collections.abc.Mapping):
        for key, value in source.items():
            if isinstance(value, collections.abc.Mapping):
                target[key] = deep_update(target.get(key, {}), cast(DeepUpdateType, value))
            else:
                target[key] = value
    else:
        target = copy.copy(source)
    return target



#
# Copyright 2019 Amazon.com, Inc. or its affiliates. All Rights Reserved.
# Copyright (C) 2018-2019  OpenCyphal Development Team  <opencyphal.org>
# This software is distributed under the terms of the MIT License.
#
"""
Module containing types and utilities for building generator objects.
Generators abstract the code generation technology used to transform
pydsdl AST into source code.
"""

import abc
import pathlib
import typing

from pydsdl import read_namespace as read_dsdl_namespace

from nunavut._namespace import Namespace, build_namespace_tree
from nunavut._utilities import YesNoDefault
from nunavut.lang import LanguageContextBuilder
from nunavut.lang._language import Language


class AbstractGenerator(metaclass=abc.ABCMeta):
    """
    Abstract base class for classes that generate source file output
    from a given pydsdl parser result.

    :param nunavut.Namespace namespace:  The top-level namespace to
        generates types at and from.
    :param YesNoDefault generate_namespace_types:  Set to YES
        to force generation files for namespaces and NO to suppress.
        DEFAULT will generate namespace files based on the language
        preference.
    """

    def __init__(
        self,
        namespace: Namespace,
        generate_namespace_types: YesNoDefault = YesNoDefault.DEFAULT,
    ):
        self._namespace = namespace
        if generate_namespace_types == YesNoDefault.YES:
            self._generate_namespace_types = True
        elif generate_namespace_types == YesNoDefault.NO:
            self._generate_namespace_types = False
        else:
            target_language = self._namespace.get_language_context().get_target_language()
            if target_language.has_standard_namespace_files:
                self._generate_namespace_types = True
            else:
                self._generate_namespace_types = False

    @property
    def namespace(self) -> Namespace:
        """
        The root :class:`nunavut.Namespace` for this generator.
        """
        return self._namespace

    @property
    def generate_namespace_types(self) -> bool:
        """
        If true then the generator is set to emit files for :class:`nunavut.Namespace`
        in addition to the pydsdl datatypes. If false then only files for pydsdl datatypes
        will be generated.
        """
        return self._generate_namespace_types

    @abc.abstractmethod
    def get_templates(self, omit_serialization_support: bool = False) -> typing.Iterable[pathlib.Path]:
        """
        Enumerate all templates found in the templates path.
        :param bool omit_serialization_support: If True then templates needed only for serialization will be omitted.
        :return: A list of paths to all templates found by this Generator object.
        """
        raise NotImplementedError()

    @abc.abstractmethod
    def generate_all(
        self, is_dryrun: bool = False, allow_overwrite: bool = True, omit_serialization_support: bool = False
    ) -> typing.Iterable[pathlib.Path]:
        """
        Generates all output for a given :class:`nunavut.Namespace` and using
        the templates found by this object.

        :param bool is_dryrun: If True then no output files will actually be
                               written but all other operations will be performed.
        :param bool allow_overwrite: If True then the generator will attempt to overwrite any existing files
                                it encounters. If False then the generator will raise an error if the
                                output file exists and the generation is not a dry-run.
        :param bool omit_serialization_support: If True then the generator will emit only types without additional
                                serialization and deserialization support and logic.
        :return: 0 for success. Non-zero for errors.
        :raises: PermissionError if :attr:`allow_overwrite` is False and the file exists.
        """
        raise NotImplementedError()


def create_default_generators(
    namespace: Namespace, **kwargs: typing.Any
) -> typing.Tuple["AbstractGenerator", "AbstractGenerator"]:
    """
    Create the two generators used by Nunavut; a code-generator and a support-library generator.

    :param nunavut.Namespace namespace: The namespace to generate code within.
    :param kwargs: A list of arguments that are forwarded to the generator constructors.
    :return: Tuple with the first item being the code-generator and the second the support-library
        generator.
    """
    from nunavut.jinja import DSDLCodeGenerator, SupportGenerator

    return (DSDLCodeGenerator(namespace, **kwargs), SupportGenerator(namespace, **kwargs))


# +---------------------------------------------------------------------------+
# | GENERATION HELPERS
# +---------------------------------------------------------------------------+


def generate_types(
    language_key: str,
    root_namespace_dir: pathlib.Path,
    out_dir: pathlib.Path,
    omit_serialization_support: bool = True,
    is_dryrun: bool = False,
    allow_overwrite: bool = True,
    lookup_directories: typing.Optional[typing.Iterable[str]] = None,
    allow_unregulated_fixed_port_id: bool = False,
    language_options: typing.Mapping[str, typing.Any] = {},
    include_experimental_languages: bool = False,
) -> None:
    """
    Helper method that uses default settings and built-in templates to generate types for a given
    language. This method is the most direct way to generate code using Nunavut.

    :param str language_key: The name of the language to generate source for.
                See the :doc:`../../docs/templates` for details on available language support.
    :param pathlib.Path root_namespace_dir: The path to the root of the DSDL types to generate
                code for.
    :param pathlib.Path out_dir: The path to generate code at and under.
    :param bool omit_serialization_support: If True then logic used to serialize and deserialize data is omitted.
    :param bool is_dryrun: If True then nothing is generated but all other activity is performed and any errors
                that would have occurred are reported.
    :param bool allow_overwrite: If True then generated files are allowed to overwrite existing files under the
                `out_dir` path.
    :param typing.Optional[typing.Iterable[str]] lookup_directories: Additional directories to search for dependent
                types referenced by the types provided under the `root_namespace_dir`. Types will not be generated
                for these unless they are used by a type in the root namespace.
    :param bool allow_unregulated_fixed_port_id: If True then errors will become warning when using fixed port
                identifiers for unregulated datatypes.
    :param typing.Optional[typing.Mapping[str, typing.Any]] language_options: Opaque arguments passed through to the
                language objects. The supported arguments and valid values are different depending on the language
                specified by the `language_key` parameter.
    :param bool include_experimental_languages: If true then experimental languages will also be available.
    """

    language_context = (
        LanguageContextBuilder(include_experimental_languages=include_experimental_languages)
        .set_target_language(language_key)
        .set_target_language_configuration_override(Language.WKCV_LANGUAGE_OPTIONS, language_options)
        .create()
    )

    if lookup_directories is None:
        lookup_directories = []

    type_map = read_dsdl_namespace(
        str(root_namespace_dir), lookup_directories, allow_unregulated_fixed_port_id=allow_unregulated_fixed_port_id
    )

    namespace = build_namespace_tree(type_map, str(root_namespace_dir), str(out_dir), language_context)

    generator, support_generator = create_default_generators(namespace)
    support_generator.generate_all(is_dryrun, allow_overwrite, omit_serialization_support)
    generator.generate_all(is_dryrun, allow_overwrite, omit_serialization_support)



#
# Copyright 2022 Amazon.com, Inc. or its affiliates. All Rights Reserved.
# Copyright (C) 2018-2022  OpenCyphal Development Team  <opencyphal.org>
# This software is distributed under the terms of the MIT License.
#
"""
Code generator built on top of pydsdl.

Nunavut uses pydsdl to generate text files using templates. While these
text files are often source code this module could also be used to generate
documentation or data interchange formats like JSON or XML.

The input to the nunavut library is a list of templates and a list of
``pydsdl.pydsdl.CompositeType`` objects. The latter is typically obtained
by calling pydsdl::

    from pydsdl import read_namespace

    compound_types = read_namespace(root_namespace, include_paths)

Next a :class:`nunavut.LanguageContext` is needed which is used to
configure all Nunavut objects for a specific target language

    .. code-block:: python

        from nunavut import LanguageContextBuilder

        # Here we are going to generate C headers.
        language_context = LanguageContextBuilder().set_target_language("c").create()

:class:`nunavut.AbstractGenerator` objects require
a :class:`nunavut.Namespace` tree which can be built from the
pydsdl type map using :meth:`nunavut.build_namespace_tree`::

    from nunavut import build_namespace_tree

    root_namespace = build_namespace_tree(compound_types,
                                          root_ns_folder,
                                          out_dir,
                                          language_context)

Putting this all together, the typical use of this library looks something like this::

    from pydsdl import read_namespace
    from nunavut import build_namespace_tree
    from nunavut.lang import LanguageContextBuilder
    from nunavut.jinja import DSDLCodeGenerator

    # parse the dsdl
    compound_types = read_namespace(root_namespace, include_paths)

    # select a target language
    language_context = LanguageContextBuilder().set_target_language("c").create()


    # build the namespace tree
    root_namespace = build_namespace_tree(compound_types,
                                          root_ns_folder,
                                          out_dir,
                                          language_context)

    # give the root namespace to the generator and...
    generator = DSDLCodeGenerator(root_namespace)

    # generate all the code!
    generator.generate_all()

"""
import sys as _sys

from ._generators import AbstractGenerator as AbstractGenerator
from ._generators import generate_types as generate_types
from ._namespace import Namespace as Namespace
from ._namespace import build_namespace_tree as build_namespace_tree
from ._utilities import TEMPLATE_SUFFIX as TEMPLATE_SUFFIX
from ._utilities import YesNoDefault
from ._version import __author__ as __author__
from ._version import __copyright__ as __copyright__
from ._version import __email__ as __email__
from ._version import __license__ as __license__
from ._version import __version__ as __version__
from .jinja import CodeGenerator as CodeGenerator
from .jinja import DSDLCodeGenerator as DSDLCodeGenerator
from .jinja import SupportGenerator as SupportGenerator
from .lang import Language as Language
from .lang import LanguageContext as LanguageContext
from .lang import LanguageContextBuilder as LanguageContextBuilder
from .lang import UnsupportedLanguageError as UnsupportedLanguageError
from .lang._config import LanguageConfig as LanguageConfig
from ._exceptions import InternalError as InternalError

if _sys.version_info[:2] < (3, 5):  # pragma: no cover
    print("A newer version of Python is required", file=_sys.stderr)
    _sys.exit(1)

__version_info__ = tuple(map(int, __version__.split(".")[:3]))

__all__ = [
    "AbstractGenerator",
    "build_namespace_tree",
    "CodeGenerator",
    "DSDLCodeGenerator",
    "generate_types",
    "LanguageConfig",
    "Language",
    "LanguageContext",
    "LanguageContextBuilder",
    "Namespace",
    "SupportGenerator",
    "TEMPLATE_SUFFIX",
    "UnsupportedLanguageError",
    "YesNoDefault",
    "InternalError",
    "__author__",
    "__copyright__",
    "__email__",
    "__license__",
    "__version__",
    "__version_info__",
]



#
# Copyright 2021 Amazon.com, Inc. or its affiliates. All Rights Reserved.
# Copyright (C) 2018-2021  OpenCyphal Development Team  <opencyphal.org>
# This software is distributed under the terms of the MIT License.
#
"""
    Objects that utilize command-line inputs to run a program using Nunavut.
"""
import argparse
import pathlib
import sys
import typing

from pydsdl import read_namespace as read_dsdl_namespace

from nunavut._generators import create_default_generators
from nunavut._namespace import build_namespace_tree
from nunavut._postprocessors import (
    ExternalProgramEditInPlace,
    FilePostProcessor,
    LimitEmptyLines,
    PostProcessor,
    SetFileMode,
    TrimTrailingWhitespace,
)
from nunavut._utilities import YesNoDefault
from nunavut.lang import Language, LanguageContext, LanguageContextBuilder


class ArgparseRunner:
    """
    Runner that uses Python argparse arguments to define a run.

    :param argparse.Namespace args: The commandline arguments.
    :param typing.Optional[typing.Union[str, typing.List[str]]] extra_includes: A list of paths to additional DSDL
        root folders.
    """

    def __init__(self, args: argparse.Namespace, extra_includes: typing.Optional[typing.Union[str, typing.List[str]]]):
        self._args = args

        if extra_includes is None:
            extra_includes = []
        elif not isinstance(extra_includes, list):
            extra_includes = [extra_includes]

        self._extra_includes = extra_includes

        #
        # nunavut : parse inputs
        #
        self._language_context = self._create_language_context()

        if self._args.generate_support != "only" and not self._args.list_configuration:
            type_map = read_dsdl_namespace(
                self._args.root_namespace,
                self._extra_includes,
                allow_unregulated_fixed_port_id=self._args.allow_unregulated_fixed_port_id,
            )
        else:
            type_map = []

        self._root_namespace = build_namespace_tree(
            type_map, self._args.root_namespace, self._args.outdir, self._language_context
        )

        #
        # nunavut : create generators
        #

        generator_args = {
            "generate_namespace_types": (
                YesNoDefault.YES if self._args.generate_namespace_types else YesNoDefault.DEFAULT
            ),
            "templates_dir": (pathlib.Path(self._args.templates) if self._args.templates is not None else None),
            "trim_blocks": self._args.trim_blocks,
            "lstrip_blocks": self._args.lstrip_blocks,
            "post_processors": self._build_post_processor_list_from_args(),
        }

        self._generator, self._support_generator = create_default_generators(self._root_namespace, **generator_args)

    def run(self) -> None:
        """
        Perform actions defined by the arguments this object was created with. This may generate outputs where
        the arguments have requested this action.

        .. warning::
            :meth:`setup` must be called before calling this method.

        """

        if self._args.list_outputs:
            self._list_outputs_only()

        elif self._args.list_inputs:
            self._list_inputs_only()

        elif self._args.list_configuration:
            self._list_configuration_only()

        else:
            self._generate()

    # +---------------------------------------------------------------------------------------------------------------+
    # | PRIVATE
    # +---------------------------------------------------------------------------------------------------------------+

    def _should_generate_support(self) -> bool:
        if self._args.generate_support == "as-needed":
            return self._args.omit_serialization_support is None or not self._args.omit_serialization_support
        else:
            return bool(self._args.generate_support == "always" or self._args.generate_support == "only")

    def _build_ext_program_postprocessor(self, program: str) -> FilePostProcessor:
        subprocess_args = [program]
        if hasattr(self._args, "pp_run_program_arg") and self._args.pp_run_program_arg is not None:
            for program_arg in self._args.pp_run_program_arg:
                subprocess_args.append(program_arg)
        return ExternalProgramEditInPlace(subprocess_args)

    def _build_post_processor_list_from_args(self) -> typing.List[PostProcessor]:
        """
        Return a list of post processors setup based on the provided command-line arguments. This
        list may be empty but the function will not return None.
        """
        post_processors: typing.List[PostProcessor] = []
        if self._args.pp_trim_trailing_whitespace:
            post_processors.append(TrimTrailingWhitespace())
        if hasattr(self._args, "pp_max_emptylines") and self._args.pp_max_emptylines is not None:
            post_processors.append(LimitEmptyLines(self._args.pp_max_emptylines))
        if hasattr(self._args, "pp_run_program") and self._args.pp_run_program is not None:
            post_processors.append(self._build_ext_program_postprocessor(self._args.pp_run_program))

        post_processors.append(SetFileMode(self._args.file_mode))

        return post_processors

    def _create_language_context(self) -> LanguageContext:
        language_options = dict()
        if self._args.target_endianness is not None:
            language_options["target_endianness"] = self._args.target_endianness
        language_options["omit_float_serialization_support"] = self._args.omit_float_serialization_support
        language_options["enable_serialization_asserts"] = self._args.enable_serialization_asserts
        language_options["enable_override_variable_array_capacity"] = self._args.enable_override_variable_array_capacity
        if self._args.language_standard is not None:
            language_options["std"] = self._args.language_standard

        if self._args.configuration is None:
            additional_config_files = []
        elif isinstance(self._args.configuration, pathlib.Path):
            additional_config_files = [self._args.configuration]
        else:
            additional_config_files = self._args.configuration

        target_language_name = self._args.target_language

        return (
            LanguageContextBuilder(include_experimental_languages=self._args.experimental_languages)
            .set_target_language(target_language_name)
            .set_additional_config_files(additional_config_files)
            .set_target_language_extension(self._args.output_extension)
            .set_target_language_configuration_override(
                Language.WKCV_NAMESPACE_FILE_STEM, self._args.namespace_output_stem
            )
            .set_target_language_configuration_override(Language.WKCV_LANGUAGE_OPTIONS, language_options)
            .create()
        )

    # +---------------------------------------------------------------------------------------------------------------+
    # | PRIVATE :: RUN METHODS
    # +---------------------------------------------------------------------------------------------------------------+
    def _stdout_lister(
        self, things_to_list: typing.Iterable[typing.Any], to_string: typing.Callable[[typing.Any], str]
    ) -> None:
        for thing in things_to_list:
            sys.stdout.write(to_string(thing))
            sys.stdout.write(";")

    def _list_outputs_only(self) -> None:
        if self._args.generate_support != "only":
            self._stdout_lister(self._generator.generate_all(is_dryrun=True), lambda p: str(p))

        if self._should_generate_support():
            self._stdout_lister(self._support_generator.generate_all(is_dryrun=True), lambda p: str(p))

    def _list_inputs_only(self) -> None:
        if self._args.generate_support != "only":
            self._stdout_lister(
                self._generator.get_templates(omit_serialization_support=self._args.omit_serialization_support),
                lambda p: str(p.resolve()),
            )

        if self._should_generate_support():
            self._stdout_lister(
                self._support_generator.get_templates(omit_serialization_support=self._args.omit_serialization_support),
                lambda p: str(p.resolve()),
            )

        if self._args.generate_support != "only":
            if self._generator.generate_namespace_types:
                self._stdout_lister(
                    [x for x, _ in self._root_namespace.get_all_types()], lambda p: str(p.source_file_path.as_posix())
                )
            else:
                self._stdout_lister(
                    [x for x, _ in self._root_namespace.get_all_datatypes()],
                    lambda p: str(p.source_file_path.as_posix()),
                )

    def _list_configuration_only(self) -> None:
        lctx = self._language_context

        import yaml

        sys.stdout.write("target_language: '")
        sys.stdout.write(lctx.get_target_language().name)
        sys.stdout.write("'\n")

        yaml.dump(lctx.config.sections(), sys.stdout, allow_unicode=True)

    def _generate(self) -> None:
        if self._should_generate_support():
            self._support_generator.generate_all(
                is_dryrun=self._args.dry_run,
                allow_overwrite=not self._args.no_overwrite,
                omit_serialization_support=self._args.omit_serialization_support,
            )

        if self._args.generate_support != "only":
            self._generator.generate_all(
                is_dryrun=self._args.dry_run,
                allow_overwrite=not self._args.no_overwrite,
                omit_serialization_support=self._args.omit_serialization_support,
            )



#
# Copyright 2023 Amazon.com, Inc. or its affiliates. All Rights Reserved.
# Copyright (C) 2023  OpenCyphal Development Team  <opencyphal.org>
# This software is distributed under the terms of the MIT License.
#
"""
Exception types thrown by the core library.

"""


class InternalError(RuntimeError):
    """Internal, opaque error within Nunavut.

    This exception is a "should never happen" exception. If caught you've probably hit a bug.
    This is the only exception type within the library that can be use where no unit tests are covering the error
    (i.e. pragma: no cover branches).
    """
